from typing import List

import hypothesis.strategies as st
from hypothesis import given, settings

from pysmooth.__main__ import smooth


@settings(deadline=None)
@given(x=st.lists(elements=st.integers(), min_size=3))
def test_hypo_3RS3R_int(x: List[float]) -> None:
    smooth(x=x, kind="3RS3R", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3RS3R", twiceit=False, endrule="copy")

    smooth(x=x, kind="3RS3R", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3RS3R", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.integers(), min_size=3))
def test_hypo_3RSS_int(x: List[float]) -> None:
    smooth(x=x, kind="3RSS", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3RSS", twiceit=False, endrule="copy")

    smooth(x=x, kind="3RSS", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3RSS", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.integers(), min_size=3))
def test_hypo_3RSR_int(x: List[float]) -> None:
    smooth(x=x, kind="3RSR", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3RSR", twiceit=False, endrule="copy")

    smooth(x=x, kind="3RSR", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3RSR", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.integers(), min_size=3))
def test_hypo_3R_int(x: List[float]) -> None:
    smooth(x=x, kind="3R", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3R", twiceit=False, endrule="copy")

    smooth(x=x, kind="3R", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3R", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.integers(), min_size=3))
def test_hypo_3_int(x: List[float]) -> None:
    smooth(x=x, kind="3", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3", twiceit=False, endrule="copy")

    smooth(x=x, kind="3", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.integers(), min_size=2, max_size=2))
def test_hypo_3_2(x: List[float]) -> None:
    smooth(x=x, kind="3", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3", twiceit=False, endrule="copy")

    smooth(x=x, kind="3", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.integers(), min_size=3))
def test_hypo_S_int(x: List[float]) -> None:
    smooth(x=x, kind="S", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="S", twiceit=False, endrule="copy")

    smooth(x=x, kind="S", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="S", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.floats(allow_nan=False, allow_infinity=False), min_size=3))
def test_hypo_3RS3R_float(x: List[float]) -> None:
    smooth(x=x, kind="3RS3R", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3RS3R", twiceit=False, endrule="copy")

    smooth(x=x, kind="3RS3R", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3RS3R", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.floats(allow_nan=False, allow_infinity=False), min_size=3))
def test_hypo_3RSS_float(x: List[float]) -> None:
    smooth(x=x, kind="3RSS", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3RSS", twiceit=False, endrule="copy")

    smooth(x=x, kind="3RSS", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3RSS", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.floats(allow_nan=False, allow_infinity=False), min_size=3))
def test_hypo_3RSR_float(x: List[float]) -> None:
    smooth(x=x, kind="3RSR", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3RSR", twiceit=False, endrule="copy")

    smooth(x=x, kind="3RSR", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3RSR", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.floats(allow_nan=False, allow_infinity=False), min_size=3))
def test_hypo_3R_float(x: List[float]) -> None:
    smooth(x=x, kind="3R", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3R", twiceit=False, endrule="copy")

    smooth(x=x, kind="3R", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3R", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.floats(allow_nan=False, allow_infinity=False), min_size=3))
def test_hypo_3_float(x: List[float]) -> None:
    smooth(x=x, kind="3", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="3", twiceit=False, endrule="copy")

    smooth(x=x, kind="3", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="3", twiceit=True, endrule="copy")


@settings(deadline=None)
@given(x=st.lists(elements=st.floats(allow_nan=False, allow_infinity=False), min_size=3))
def test_hypo_S_float(x: List[float]) -> None:
    smooth(x=x, kind="S", twiceit=False, endrule="Tukey")
    smooth(x=x, kind="S", twiceit=False, endrule="copy")

    smooth(x=x, kind="S", twiceit=True, endrule="Tukey")
    smooth(x=x, kind="S", twiceit=True, endrule="copy")



"""Command-line interface."""

import numpy as np

from pysmooth._smooth import sm_3, sm_3R, sm_3RS3R, sm_3RSR, sm_3RSS, sm_split3


def smooth(
    x: list[float],
    kind: str = "3RS3R",
    twiceit: bool = False,
    endrule: str = "Tukey",
    do_ends: bool = False,
) -> list[float]:
    """Tukey's smoothers, translated from the `smooth` function found in the R module {stats}.

    Parameters
    ----------
    x : arraylike
        a list of numbers to smooth
    kind : str, optional
        the kind of smoother required.  Valid options include "3RS3R", "3RSS",
        "3RSR", "3R", "3", "S"
    twiceit : bool, optional
        Run the smoother over the data twice.  Twicing a smoother S(y) means
        S(y) + S(y - S(y)), i.e., adding smoothed residuals to the smoothed
        values. This decreases bias (increasing variance)
    endrule : str, optional
        rule for smoothing at the boundary. Valid options are "copy" and "Tukey"
    do_ends : bool, optional
        indicate if the 3-splitting of ties should also happen at the boundaries
        (ends). This is only used for kind = "S"

    Returns
    -------
    y : list of floats
        the smoothed values
    it : int
        number of iterations required to reach stable smoothing
    """
    if kind not in ("3RS3R", "3RSS", "3RSR", "3R", "3", "S"):
        msg = "An invalid kind of smoother was selected"
        raise RuntimeError(msg)

    if endrule == "copy":
        iend = 1
    elif endrule == "Tukey":
        iend = 2
    else:
        msg = f"endrule must be either 'copy' or 'Tukey', but {endrule} was passed"
        raise RuntimeError(msg)

    if any(np.isnan(np.array(x, dtype=np.float64))):
        msg = "The sequence to be smoothed contains NAs/nans. This algorithm " "is unable to accomodate NAs."
        raise ValueError(msg)

    if not all(np.issubdtype(type(_), np.number) for _ in x):
        msg = "The sequence contains non-float values."
        raise ValueError(msg)

    if kind.startswith("3RS") and not do_ends:
        iend = -iend
    elif kind == "S":
        iend = int(do_ends)

    n: int = len(x)
    y: list[float] = np.zeros(n, dtype=np.float64).tolist()
    if kind != "S":
        z: list[float] = np.zeros(n, dtype=np.float64).tolist()
        w: list[float] = np.zeros(n, dtype=np.float64).tolist()
        split_ends = iend < 0

        if kind == "3RS3R":
            _, x, y, z, w = sm_3RS3R(x, y, z, w, n, abs(iend), split_ends)
        elif kind == "3RSS":
            _, x, y, z = sm_3RSS(x, y, z, n, abs(iend), split_ends)
        elif kind == "3RSR":
            _, x, y, z, w = sm_3RSR(x, y, z, w, n, abs(iend), split_ends)
        elif kind == "3R":
            _, x, y, z = sm_3R(x, y, z, n, iend)
        elif kind == "3":
            _, x, y = sm_3(x, y, n, iend)

    else:
        _, x, y = sm_split3(x, y, n, bool(iend))

    if twiceit:
        r: list[float] = smooth(x=y, kind=kind, twiceit=False, endrule=endrule, do_ends=do_ends)
        y += r

    return y



"""Tukey's (Running Median) smoother.

A direct as I can translation of the R and C code for stats::smooth() found
in R 4.1.2.  A copy of the original code can be found at
https://github.com/wch/r-source/blob/trunk/src/library/stats/src/smooth.c

"""

from enum import Enum
from statistics import median


class EndRule(int, Enum):
    sm_no_endrule = 0
    sm_copy_endrule = 1
    sm_tukey_endrule = 2


N_CUTOFF = 2


# TODO: reimplement this in rust
def med3(u: float, v: float, w: float) -> float:
    """Find the median of three numbers.

    Parameters
    ----------
    u, v, w : float
        The three numbers to find the median of.

    Returns
    -------
    float
        median of u, v, w
    """
    if u <= v <= w or u >= v >= w:
        return v
    elif u <= w <= v or u >= w >= v:
        return w
    else:
        return u


def imed3(u: float, v: float, w: float) -> int:
    """Find the index of the median of three numbers.

    Parameters
    ----------
    u, v, w : float
        The three numbers to find the index of the median of.

    Returns
    -------
    int
        -1 if u, 0 if v, or 1 if w is the median
    """
    if u <= v <= w or u >= v >= w:
        return 0
    elif u <= w <= v or u >= w >= v:
        return 1
    else:
        return -1


def sm_3(x: list[float], y: list[float], n: int, end_rule: int) -> tuple[bool, list[float], list[float]]:
    chg = False

    if n <= N_CUTOFF:
        for i in range(n):
            y[i] = x[i]
        return False, x, y

    for i in range(1, n - 1):
        j = imed3(x[i - 1], x[i], x[i + 1])
        y[i] = x[i + j]
        chg = chg or bool(j)

    chg, y, x = sm_do_endrule(y, x, n, chg, end_rule)

    return chg, x, y


def sm_do_endrule(
    y: list[float],
    x: list[float],
    n: int,
    chg: bool,
    end_rule: int,
) -> tuple[bool, list[float], list[float]]:
    if end_rule == EndRule.sm_no_endrule:
        pass
    elif end_rule == EndRule.sm_copy_endrule:
        y[0] = x[0]
        y[n - 1] = x[n - 1]
    elif end_rule == EndRule.sm_tukey_endrule:
        y[0] = median([3 * y[1] - 2 * y[2], x[0], y[1]])
        chg = chg or (y[0] != x[0])
        y[n - 1] = median([y[n - 2], x[n - 1], 3 * y[n - 2] - 2 * y[n - 3]])
        chg = chg or (y[n - 1] != x[n - 1])
    else:
        msg = f"invalid end-rule for running median of 3: {end_rule}"
        raise RuntimeError(msg)

    return chg, y, x


def sm_3R(
    x: list[float],
    y: list[float],
    z: list[float],
    n: int,
    end_rule: int,
) -> tuple[int, list[float], list[float], list[float]]:
    chg, x, y = sm_3(x, y, n, 1)  # "sm_COPY_ENDRULE"
    it = int(chg)

    while chg:
        chg, y, z = sm_3(y, z, n, 0)  # "sm_NO_ENDRULE"
        if chg:
            it += 1
            for i in range(1, n - 1):
                y[i] = z[i]

    if n > 2:
        chg, y, x = sm_do_endrule(y, x, n, chg, end_rule)

    return (it, x, y, z) if it else (int(chg), x, y, z)


def sptest(x: list[float], i: int) -> bool:
    if x[i] != x[i + 1]:
        return False
    elif (x[i - 1] <= x[i] and x[i + 1] <= x[i + 2]) or (x[i - 1] >= x[i] and x[i + 1] >= x[i + 2]):
        return False
    else:
        return True


def sm_split3(x: list[float], y: list[float], n: int, do_ends: bool) -> tuple[bool, list[float], list[float]]:
    # y[] := S(x[])  where S() = "sm_split3"
    chg = False

    for i in range(n):
        y[i] = x[i]

    if n <= 4:
        return False, y, x

    # Colin Goodall doesn't do splits near ends
    # in spl() in Statlib's "smoother" code !!
    if do_ends and sptest(x, 1):
        chg = True
        y[1] = x[0]
        y[2] = med3(x[2], x[3], 3 * x[3] - 2 * x[4])

    for i in range(2, n - 3):
        if sptest(x, i):
            # plateau at x[i] == x[i+1]
            # at left :
            if (j := imed3(x[i], x[i - 1], 3 * x[i - 1] - 2 * x[i - 2])) > -1:
                y[i] = x[i - 1] if j == 0 else 3 * x[i - 1] - 2 * x[i - 2]
                chg = y[i] != x[i]
            # at right :
            if (j := imed3(x[i + 1], x[i + 2], 3 * x[i + 2] - 2 * x[i + 3])) > -1:
                y[i + 1] = x[i + 2] if j == 0 else 3 * x[i + 2] - 2 * x[i + 3]
                chg = y[i + 1] != x[i + 1]
    if do_ends and sptest(x, n - 3):
        chg = True
        y[n - 2] = x[n - 1]
        y[n - 3] = med3(x[n - 3], x[n - 4], 3 * x[n - 4] - 2 * x[n - 5])
    return chg, x, y


def sm_3RS3R(
    x: list[float],
    y: list[float],
    z: list[float],
    w: list[float],
    n: int,
    end_rule: int,
    split_ends: bool,
) -> tuple[int, list[float], list[float], list[float], list[float]]:
    # y[1:n] := "3R S 3R"(x[1:n]);  z = "work";

    it, x, y, z = sm_3R(x, y, z, n, end_rule)
    chg, y, z = sm_split3(y, z, n, split_ends)
    if chg:
        (
            it2,
            z,
            y,
            w,
        ) = sm_3R(z, y, w, n, end_rule)
        it += it2
    # else y == z already
    return (it + chg), x, y, z, w


def sm_3RSS(
    x: list[float],
    y: list[float],
    z: list[float],
    n: int,
    end_rule: int,
    split_ends: bool,
) -> tuple[int, list[float], list[float], list[float]]:
    # y[1:n] := "3RSS"(x[1:n]);  z = "work"

    it, x, y, z = sm_3R(x, y, z, n, end_rule)
    chg, y, z = sm_split3(y, z, n, split_ends)
    if chg is True:
        chg, z, y = sm_split3(z, y, n, split_ends)
    # else  y == z already
    return (it + int(chg)), x, y, z


def sm_3RSR(
    x: list[float],
    y: list[float],
    z: list[float],
    w: list[float],
    n: int,
    end_rule: int,
    split_ends: bool,
) -> tuple[int, list[float], list[float], list[float], list[float]]:
    # y[1:n] := "3RSR"(x[1:n]);  z := residuals; w = "work";

    # == "SR" (as follows) is stupid ! (MM) ==

    it, x, y, z = sm_3R(x, y, z, n, end_rule)
    chg: bool = True
    while chg:
        it += 1
        chg, y, z = sm_split3(y, z, n, split_ends)
        ch2, z, y, w = sm_3R(z, y, w, n, end_rule)
        chg = chg or bool(ch2)

        if chg is False:
            break
        if it > 2 * n:
            break  # INF.LOOP stopper
        for i in range(n):
            z[i] = x[i] - y[i]

    return it, x, y, z, w
